"""Color palettes for your palate"""

from .core import cook_tmap, pair_tbar
from .models import TastyBar, TastyMap
from .ui import TastyKitchen

__version__ = "0.2.0"

__all__ = ["cook_tmap", "pair_tbar", "TastyMap", "TastyBar", "TastyKitchen"]
